@@ -37,7 +37,8 @@ if not PY3K:
     import io
 # Fortran statement matching rules
 USE_REGEX = re.compile(
-    r"[ ]*USE([, ]+INTRINSIC)?[ :]+([a-z0-9_]*)([, ]+ONLY[ :]+)?", re.I
+    r"[ ]*USE([, ]+(?:INTRINSIC|NON_INTRINSIC))?[ :]+(\w*)([, ]+ONLY[ :]+)?",
+    re.I,
 )
 IMPORT_REGEX = re.compile(r"[ ]*IMPORT[ :]+([a-z_])", re.I)
 INCLUDE_REGEX = re.compile(r"[ ]*INCLUDE[ :]*[\'\"]([^\'\"]*)", re.I)
@@ -693,7 +694,7 @@ def read_submod_def(line):
             else:
                 trailing_line = ""
         #
-        name_match = WORD_REGEX.match(trailing_line)
+        name_match = WORD_REGEX.search(trailing_line)
         if name_match is not None:
             name = name_match.group(0).lower()
         return "smod", SMOD_info(name, parent_name)
